/*
 * Copyright (C) 2006-2013 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file fd.c Helper functions for file descriptor handling. */

#include <regex.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <sys/uio.h>

#include "para.h"
#include "error.h"
#include "string.h"
#include "fd.h"

/**
 * Write an array of buffers to a file descriptor.
 *
 * \param fd The file descriptor.
 * \param iov Pointer to one or more buffers.
 * \param iovcnt The number of buffers.
 *
 * EAGAIN/EWOULDBLOCK is not considered a fatal error condition. For example
 * DCCP CCID3 has a sending wait queue which fills up and is emptied
 * asynchronously. The EAGAIN case means that there is currently no space in
 * the wait queue, but this can change at any moment.
 *
 * \return Negative on fatal errors, number of bytes written else.
 *
 * For blocking file descriptors, this function returns either the sum of all
 * buffer sizes, or the error code of the fatal error that caused the last
 * write call to fail.
 *
 * For nonblocking file descriptors there is a third possibility: Any positive
 * return value less than the sum of the buffer sizes indicates that some bytes
 * have been written but the next write would block.
 *
 * \sa writev(2), \ref xwrite().
 */
int xwritev(int fd, struct iovec *iov, int iovcnt)
{
	size_t written = 0;
	int i;
	struct iovec saved_iov, *curiov;

	i = 0;
	curiov = iov;
	saved_iov = *curiov;
	while (i < iovcnt && curiov->iov_len > 0) {
		ssize_t ret = writev(fd, curiov, iovcnt - i);
		if (ret >= 0) {
			written += ret;
			while (ret > 0) {
				if (ret < curiov->iov_len) {
					curiov->iov_base += ret;
					curiov->iov_len -= ret;
					break;
				}
				ret -= curiov->iov_len;
				*curiov = saved_iov;
				i++;
				if (i >= iovcnt)
					return written;
				curiov++;
				saved_iov = *curiov;
			}
			continue;
		}
		if (errno == EINTR)
			/*
			 * The write() call was interrupted by a signal before
			 * any data was written. Try again.
			 */
			continue;
		if (errno == EAGAIN || errno == EWOULDBLOCK)
			/*
			 * We don't consider this an error. Note that POSIX
			 * allows either error to be returned, and does not
			 * require these constants to have the same value.
			 */
			return written;
		/* fatal error */
		return -ERRNO_TO_PARA_ERROR(errno);
	}
	return written;
}

/**
 * Write a buffer to a file descriptor, re-writing on short writes.
 *
 * \param fd The file descriptor.
 * \param buf The buffer to write.
 * \param len The number of bytes to write.
 *
 * This is a simple wrapper for \ref xwritev().
 *
 * \return The return value of the underlying call to \ref xwritev().
 */
int xwrite(int fd, const char *buf, size_t len)
{
	struct iovec iov = {.iov_base = (void *)buf, .iov_len = len};
	return xwritev(fd, &iov, 1);
}

/**
 * Write all data to a file descriptor.
 *
 * \param fd The file descriptor.
 * \param buf The buffer to be sent.
 * \param len The length of \a buf.
 *
 * This is like \ref xwrite() but returns \p -E_SHORT_WRITE if not
 * all data could be written.
 *
 * \return Number of bytes written on success, negative error code else.
 */
int write_all(int fd, const char *buf, size_t len)
{
	int ret = xwrite(fd, buf, len);

	if (ret < 0)
		return ret;
	if (ret != len)
		return -E_SHORT_WRITE;
	return ret;
}

/**
 * Write a buffer given by a format string.
 *
 * \param fd The file descriptor.
 * \param fmt A format string.
 *
 * \return The return value of the underlying call to \ref write_all().
 */
__printf_2_3 int write_va_buffer(int fd, const char *fmt, ...)
{
	char *msg;
	int ret;
	va_list ap;

	va_start(ap, fmt);
	ret = xvasprintf(&msg, fmt, ap);
	va_end(ap);
	ret = write_all(fd, msg, ret);
	free(msg);
	return ret;
}

/**
 * Read from a non-blocking file descriptor into multiple buffers.
 *
 * \param fd The file descriptor to read from.
 * \param iov Scatter/gather array used in readv().
 * \param iovcnt Number of elements in \a iov.
 * \param rfds An optional fd set pointer.
 * \param num_bytes Result pointer. Contains the number of bytes read from \a fd.
 *
 * If \a rfds is not \p NULL and the (non-blocking) file descriptor \a fd is
 * not set in \a rfds, this function returns early without doing anything.
 * Otherwise The function tries to read up to \a sz bytes from \a fd, where \a
 * sz is the sum of the lengths of all vectors in \a iov. As for xwrite(),
 * \p EAGAIN is not considered an error condition. However, \p EOF is.
 *
 * \return Zero or a negative error code. If the underlying call to readv(2)
 * returned zero (indicating an end of file condition) or failed for some
 * reason other than \p EAGAIN, a negative return value is returned.
 *
 * In any case, \a num_bytes contains the number of bytes that have been
 * successfully read from \a fd (zero if the first readv() call failed with
 * EAGAIN). Note that even if the function returns negative, some data might
 * have been read before the error occurred. In this case \a num_bytes is
 * positive.
 *
 * \sa \ref xwrite(), read(2), readv(2).
 */
int readv_nonblock(int fd, struct iovec *iov, int iovcnt, fd_set *rfds,
		size_t *num_bytes)
{
	int ret, i, j;

	*num_bytes = 0;
	/*
	 * Avoid a shortcoming of select(): Reads from a non-blocking fd might
	 * return EAGAIN even if FD_ISSET() returns true. However, FD_ISSET()
	 * returning false definitely means that no data can currently be read.
	 * This is the common case, so it is worth to avoid the overhead of the
	 * read() system call in this case.
	 */
	if (rfds && !FD_ISSET(fd, rfds))
		return 0;

	for (i = 0, j = 0; i < iovcnt;) {

		/* fix up the first iov */
		assert(j < iov[i].iov_len);
		iov[i].iov_base += j;
		iov[i].iov_len -= j;
		ret = readv(fd, iov + i, iovcnt - i);
		iov[i].iov_base -= j;
		iov[i].iov_len += j;

		if (ret == 0)
			return -E_EOF;
		if (ret < 0) {
			if (errno == EAGAIN)
				return 0;
			return -ERRNO_TO_PARA_ERROR(errno);
		}
		*num_bytes += ret;
		while (ret > 0) {
			if (ret < iov[i].iov_len - j) {
				j += ret;
				break;
			}
			ret -= iov[i].iov_len - j;
			j = 0;
			if (++i >= iovcnt)
				break;
		}
	}
	return 0;
}

/**
 * Read from a non-blocking file descriptor into a single buffer.
 *
 * \param fd The file descriptor to read from.
 * \param buf The buffer to read data to.
 * \param sz The size of \a buf.
 * \param rfds \see \ref readv_nonblock().
 * \param num_bytes \see \ref readv_nonblock().
 *
 * This is a simple wrapper for readv_nonblock() which uses an iovec with a single
 * buffer.
 *
 * \return The return value of the underlying call to readv_nonblock().
 */
int read_nonblock(int fd, void *buf, size_t sz, fd_set *rfds, size_t *num_bytes)
{
	struct iovec iov = {.iov_base = buf, .iov_len = sz};
	return readv_nonblock(fd, &iov, 1, rfds, num_bytes);
}

/**
 * Read a buffer and check its content for a pattern.
 *
 * \param fd The file descriptor to receive from.
 * \param pattern The expected pattern.
 * \param bufsize The size of the internal buffer.
 * \param rfds Passed to read_nonblock().
 *
 * This function tries to read at most \a bufsize bytes from the non-blocking
 * file descriptor \a fd. If at least \p strlen(\a pattern) bytes have been
 * received, the beginning of the received buffer is compared with \a pattern,
 * ignoring case.
 *
 * \return Positive if \a pattern was received, negative on errors, zero if no data
 * was available to read.
 *
 * \sa \ref read_nonblock(), \sa strncasecmp(3).
 */
int read_pattern(int fd, const char *pattern, size_t bufsize, fd_set *rfds)
{
	size_t n, len;
	char *buf = para_malloc(bufsize + 1);
	int ret = read_nonblock(fd, buf, bufsize, rfds, &n);

	buf[n] = '\0';
	if (ret < 0)
		goto out;
	ret = 0;
	if (n == 0)
		goto out;
	ret = -E_READ_PATTERN;
	len = strlen(pattern);
	if (n < len)
		goto out;
	if (strncasecmp(buf, pattern, len) != 0)
		goto out;
	ret = 1;
out:
	if (ret < 0) {
		PARA_NOTICE_LOG("%s\n", para_strerror(-ret));
		PARA_NOTICE_LOG("recvd %zu bytes: %s\n", n, buf);
	}
	free(buf);
	return ret;
}

/**
 * Check whether a file exists.
 *
 * \param fn The file name.
 *
 * \return Non-zero iff file exists.
 */
int file_exists(const char *fn)
{
	struct stat statbuf;

	return !stat(fn, &statbuf);
}

/**
 * Paraslash's wrapper for select(2).
 *
 * It calls select(2) (with no exceptfds) and starts over if select() was
 * interrupted by a signal.
 *
 * \param n The highest-numbered descriptor in any of the two sets, plus 1.
 * \param readfds fds that should be checked for readability.
 * \param writefds fds that should be checked for writablility.
 * \param timeout_tv upper bound on the amount of time elapsed before select()
 * returns.
 *
 * \return The return value of the underlying select() call on success, the
 * negative system error code on errors.
 *
 * All arguments are passed verbatim to select(2).
 * \sa select(2) select_tut(2).
 */
int para_select(int n, fd_set *readfds, fd_set *writefds,
		struct timeval *timeout_tv)
{
	int ret;
	do
		ret = select(n, readfds, writefds, NULL, timeout_tv);
	while (ret < 0 && errno == EINTR);
	if (ret < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	return ret;
}

/**
 * Set a file descriptor to blocking mode.
 *
 * \param fd The file descriptor.
 *
 * \return Standard.
 */
__must_check int mark_fd_blocking(int fd)
{
	int flags = fcntl(fd, F_GETFL);
	if (flags < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	flags = fcntl(fd, F_SETFL, ((long)flags) & ~O_NONBLOCK);
	if (flags < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	return 1;
}

/**
 * Set a file descriptor to non-blocking mode.
 *
 * \param fd The file descriptor.
 *
 * \return Standard.
 */
__must_check int mark_fd_nonblocking(int fd)
{
	int flags = fcntl(fd, F_GETFL);
	if (flags < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	flags = fcntl(fd, F_SETFL, ((long)flags) | O_NONBLOCK);
	if (flags < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	return 1;
}

/**
 * Set a file descriptor in a fd_set.
 *
 * \param fd The file descriptor to be set.
 * \param fds The file descriptor set.
 * \param max_fileno Highest-numbered file descriptor.
 *
 * This wrapper for FD_SET() passes its first two arguments to \p FD_SET. Upon
 * return, \a max_fileno contains the maximum of the old_value and \a fd.
 *
 * \sa para_select.
*/
void para_fd_set(int fd, fd_set *fds, int *max_fileno)
{
	assert(fd >= 0 && fd < FD_SETSIZE);
#if 0
	{
		int flags = fcntl(fd, F_GETFL);
		if (!(flags & O_NONBLOCK)) {
			PARA_EMERG_LOG("fd %d is a blocking file descriptor\n", fd);
			exit(EXIT_FAILURE);
		}
	}
#endif
	FD_SET(fd, fds);
	*max_fileno = PARA_MAX(*max_fileno, fd);
}

/**
 * Paraslash's wrapper for fgets(3).
 *
 * \param line Pointer to the buffer to store the line.
 * \param size The size of the buffer given by \a line.
 * \param f The stream to read from.
 *
 * \return Unlike the standard fgets() function, an integer value
 * is returned. On success, this function returns 1. On errors, -E_FGETS
 * is returned. A zero return value indicates an end of file condition.
 */
__must_check int para_fgets(char *line, int size, FILE *f)
{
again:
	if (fgets(line, size, f))
		return 1;
	if (feof(f))
		return 0;
	if (!ferror(f))
		return -E_FGETS;
	if (errno != EINTR) {
		PARA_ERROR_LOG("%s\n", strerror(errno));
		return -E_FGETS;
	}
	clearerr(f);
	goto again;
}

/**
 * Paraslash's wrapper for mmap.
 *
 * \param length Number of bytes to mmap.
 * \param prot Either PROT_NONE or the bitwise OR of one or more of
 * PROT_EXEC PROT_READ PROT_WRITE.
 * \param flags Exactly one of MAP_SHARED and MAP_PRIVATE.
 * \param fd The file to mmap from.
 * \param offset Mmap start.
 * \param map Result pointer.
 *
 * \return Standard.
 *
 * \sa mmap(2).
 */
int para_mmap(size_t length, int prot, int flags, int fd, off_t offset,
		void *map)
{
	void **m = map;

	errno = EINVAL;
	if (!length)
		goto err;
	*m = mmap(NULL, length, prot, flags, fd, offset);
	if (*m != MAP_FAILED)
		return 1;
err:
	*m = NULL;
	return -ERRNO_TO_PARA_ERROR(errno);
}

/**
 * Wrapper for the open(2) system call.
 *
 * \param path The filename.
 * \param flags The usual open(2) flags.
 * \param mode Specifies the permissions to use.
 *
 * The mode parameter must be specified when O_CREAT is in the flags, and is
 * ignored otherwise.
 *
 * \return The file descriptor on success, negative on errors.
 *
 * \sa open(2).
 */
int para_open(const char *path, int flags, mode_t mode)
{
	int ret = open(path, flags, mode);

	if (ret >= 0)
		return ret;
	return -ERRNO_TO_PARA_ERROR(errno);
}

/**
 * Wrapper for chdir(2).
 *
 * \param path The specified directory.
 *
 * \return Standard.
 */
int para_chdir(const char *path)
{
	int ret = chdir(path);

	if (ret >= 0)
		return 1;
	return -ERRNO_TO_PARA_ERROR(errno);
}

/**
 * Save the cwd and open a given directory.
 *
 * \param dirname Path to the directory to open.
 * \param dir Result pointer.
 * \param cwd File descriptor of the current working directory.
 *
 * \return Standard.
 *
 * Opening the current directory (".") and calling fchdir() to return is
 * usually faster and more reliable than saving cwd in some buffer and calling
 * chdir() afterwards.
 *
 * If \a cwd is not \p NULL "." is opened and the resulting file descriptor is
 * stored in \a cwd. If the function returns success, and \a cwd is not \p
 * NULL, the caller must close this file descriptor (probably after calling
 * fchdir(*cwd)).
 *
 * On errors, the function undos everything, so the caller needs neither close
 * any files, nor change back to the original working directory.
 *
 * \sa getcwd(3).
 *
 */
static int para_opendir(const char *dirname, DIR **dir, int *cwd)
{
	int ret;

	if (cwd) {
		ret = para_open(".", O_RDONLY, 0);
		if (ret < 0)
			return ret;
		*cwd = ret;
	}
	ret = para_chdir(dirname);
	if (ret < 0)
		goto close_cwd;
	*dir = opendir(".");
	if (*dir)
		return 1;
	ret = -ERRNO_TO_PARA_ERROR(errno);
	/* Ignore return value of fchdir() and close(). We're busted anyway. */
	if (cwd) {
		int __a_unused ret2 = fchdir(*cwd); /* STFU, gcc */
	}
close_cwd:
	if (cwd)
		close(*cwd);
	return ret;
}

/**
 * A wrapper for fchdir().
 *
 * \param fd An open file descriptor.
 *
 * \return Standard.
 */
static int para_fchdir(int fd)
{
	if (fchdir(fd) < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	return 1;
}

/**
 * A wrapper for mkdir(2).
 *
 * \param path Name of the directory to create.
 * \param mode The permissions to use.
 *
 * \return Standard.
 */
int para_mkdir(const char *path, mode_t mode)
{
	if (!mkdir(path, mode))
		return 1;
	return -ERRNO_TO_PARA_ERROR(errno);
}

/**
 * Open a file and map it into memory.
 *
 * \param path Name of the regular file to map.
 * \param open_mode Either \p O_RDONLY or \p O_RDWR.
 * \param map On success, the mapping is returned here.
 * \param size size of the mapping.
 * \param fd_ptr The file descriptor of the mapping.
 *
 * If \a fd_ptr is \p NULL, the file descriptor resulting from the underlying
 * open call is closed after mmap().  Otherwise the file is kept open and the
 * file descriptor is returned in \a fd_ptr.
 *
 * \return Standard.
 *
 * \sa para_open(), mmap(2).
 */
int mmap_full_file(const char *path, int open_mode, void **map,
		size_t *size, int *fd_ptr)
{
	int fd, ret, mmap_prot, mmap_flags;
	struct stat file_status;

	if (open_mode == O_RDONLY) {
		mmap_prot = PROT_READ;
		mmap_flags = MAP_PRIVATE;
	} else {
		mmap_prot = PROT_READ | PROT_WRITE;
		mmap_flags = MAP_SHARED;
	}
	ret = para_open(path, open_mode, 0);
	if (ret < 0)
		return ret;
	fd = ret;
	if (fstat(fd, &file_status) < 0) {
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto out;
	}
	*size = file_status.st_size;
	/*
	 * If the file is empty, *size is zero and mmap() would return EINVAL
	 * (Invalid argument). This error is common enough to spend an extra
	 * error code which explicitly states the problem.
	 */
	ret = -E_EMPTY;
	if (*size == 0)
		goto out;
	/*
	 * If fd refers to a directory, mmap() returns ENODEV (No such device),
	 * at least on Linux. "Is a directory" seems to be more to the point.
	 */
	ret = -ERRNO_TO_PARA_ERROR(EISDIR);
	if (S_ISDIR(file_status.st_mode))
		goto out;

	ret = para_mmap(*size, mmap_prot, mmap_flags, fd, 0, map);
out:
	if (ret < 0 || !fd_ptr)
		close(fd);
	else
		*fd_ptr = fd;
	return ret;
}

/**
 * A wrapper for munmap(2).
 *
 * \param start The start address of the memory mapping.
 * \param length The size of the mapping.
 *
 * \return Standard.
 *
 * \sa munmap(2), mmap_full_file().
 */
int para_munmap(void *start, size_t length)
{
	int err;

	if (!start)
		return 0;
	if (munmap(start, length) >= 0)
		return 1;
	err = errno;
	PARA_ERROR_LOG("munmap (%p/%zu) failed: %s\n", start, length,
		strerror(err));
	return -ERRNO_TO_PARA_ERROR(err);
}

/**
 * Check a file descriptor for writability.
 *
 * \param fd The file descriptor.
 *
 * \return positive if fd is ready for writing, zero if it isn't, negative if
 * an error occurred.
 */

int write_ok(int fd)
{
	struct timeval tv;
	fd_set wfds;

	FD_ZERO(&wfds);
	FD_SET(fd, &wfds);
	tv.tv_sec = 0;
	tv.tv_usec = 0;
	return para_select(fd + 1, NULL, &wfds, &tv);
}

/**
 * Ensure that file descriptors 0, 1, and 2 are valid.
 *
 * Common approach that opens /dev/null until it gets a file descriptor greater
 * than two.
 *
 * \sa okir's Black Hats Manual.
 */
void valid_fd_012(void)
{
	while (1) {
		int fd = open("/dev/null", O_RDWR);
		if (fd < 0)
			exit(EXIT_FAILURE);
		if (fd > 2) {
			close(fd);
			break;
		}
	}
}

/**
 * Traverse the given directory recursively.
 *
 * \param dirname The directory to traverse.
 * \param func The function to call for each entry.
 * \param private_data Pointer to an arbitrary data structure.
 *
 * For each regular file under \a dirname, the supplied function \a func is
 * called.  The full path of the regular file and the \a private_data pointer
 * are passed to \a func. Directories for which the calling process has no
 * permissions to change to are silently ignored.
 *
 * \return Standard.
 */
int for_each_file_in_dir(const char *dirname,
		int (*func)(const char *, void *), void *private_data)
{
	DIR *dir;
	struct dirent *entry;
	int cwd_fd, ret2, ret = para_opendir(dirname, &dir, &cwd_fd);

	if (ret < 0)
		return ret == -ERRNO_TO_PARA_ERROR(EACCES)? 1 : ret;
	/* scan cwd recursively */
	while ((entry = readdir(dir))) {
		mode_t m;
		char *tmp;
		struct stat s;

		if (!strcmp(entry->d_name, "."))
			continue;
		if (!strcmp(entry->d_name, ".."))
			continue;
		if (lstat(entry->d_name, &s) == -1)
			continue;
		m = s.st_mode;
		if (!S_ISREG(m) && !S_ISDIR(m))
			continue;
		tmp = make_message("%s/%s", dirname, entry->d_name);
		if (!S_ISDIR(m)) {
			ret = func(tmp, private_data);
			free(tmp);
			if (ret < 0)
				goto out;
			continue;
		}
		/* directory */
		ret = for_each_file_in_dir(tmp, func, private_data);
		free(tmp);
		if (ret < 0)
			goto out;
	}
	ret = 1;
out:
	closedir(dir);
	ret2 = para_fchdir(cwd_fd);
	if (ret2 < 0 && ret >= 0)
		ret = ret2;
	close(cwd_fd);
	return ret;
}
